# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright 2023 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from tikka.adapters.repository.sqlite3 import Sqlite3RepositoryInterface
from tikka.interfaces.adapters.repository.preferences import (
    PreferencesRepositoryInterface,
)

TABLE_NAME = "preferences"


class Sqlite3PreferencesRepository(
    PreferencesRepositoryInterface, Sqlite3RepositoryInterface
):
    """
    Sqlite3PreferencesRepository class
    """

    def get(self, key: str) -> Optional[str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.get.__doc__
        )

        result = self.client.select_one(
            f"SELECT value_ FROM {TABLE_NAME} WHERE key_=?", (key,)
        )
        if result is None:
            return None

        return result[0]

    def set(self, key: str, value: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.get.__doc__
        )

        # make sure it has the right data
        self.client.execute(
            f"UPDATE OR IGNORE {TABLE_NAME} SET value_=? WHERE key_='{key}';", [value]
        )

        # make sure it exists
        self.client.execute(
            f"INSERT OR IGNORE INTO {TABLE_NAME} (key_, value_) VALUES (?, ?);",
            [key, value],
        )



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from substrateinterface import Keypair

from tikka.domains.entities.technical_committee import TechnicalCommitteeProposal
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.network.technical_commitee import (
    NetworkTechnicalCommitteeInterface,
)


class TechnicalCommittee:
    """
    TechnicalCommittee domain class
    """

    def __init__(
        self,
        network: NetworkTechnicalCommitteeInterface,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init TechnicalCommittee domain

        :param network: NetworkTechnicalCommitteeInterface instance
        :param event_dispatcher: EventDispatcher instance
        """

        self.network = network
        self.event_dispatcher = event_dispatcher

    def network_members(self) -> List[str]:
        """
        Return all members of Technical Committee from network

        :return:
        """
        return self.network.members()

    def network_proposals(self) -> List[TechnicalCommitteeProposal]:
        """
        Return all proposals with voting infos of Technical Committee from network

        :return:
        """
        return self.network.proposals()

    def network_vote(
        self, keypair: Keypair, proposal: TechnicalCommitteeProposal, vote: bool
    ) -> None:
        """
        Send Technical Committee Vote for proposal from Keypair

        :param keypair: Keypair instance
        :param proposal: TechnicalCommitteeProposal instance
        :param vote: True or False
        :return:
        """
        return self.network.vote(keypair, proposal, vote)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QWidget
from substrateinterface.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.entities.constants import ADDRESS_MONOSPACE_FONT_NAME
from tikka.slots.pyqt.resources.gui.windows.address_add_rc import Ui_AddressAddDialog


class AddressAddWindow(QDialog, Ui_AddressAddDialog):
    """
    AddressAddWindow class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init add address window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        # set monospace font to address field
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressLineEdit.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.addressLineEdit.textChanged.connect(self.on_address_line_edit_changed)
        self.buttonBox.rejected.connect(self.close)

    def _normalize_address_field(self) -> Optional[DisplayAddress]:
        """
        Validate and convert address field to DisplayAddress instance

        :return:
        """
        address = self.addressLineEdit.text()
        if not is_valid_ss58_address(address):
            self.errorLabel.setText(self._("Account address is not valid!"))
            return None

        # create account instance
        account = Account(address)
        for existing_account in self.application.accounts.get_list():
            if account == existing_account:
                self.errorLabel.setText(self._("Account already exists!"))
                return None

        return DisplayAddress(address)

    def on_address_line_edit_changed(self) -> None:
        """
        Triggered when the address field is changed

        :return:
        """
        display_address = self._normalize_address_field()
        if display_address is not None:
            self.errorLabel.setText("")
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        else:
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        display_address = self._normalize_address_field()
        if display_address is not None:
            # create account instance
            account = Account(display_address.address)

            # add instance in application
            self.application.accounts.add(account)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    AddressAddWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Optional
from uuid import UUID, uuid4

from tikka.domains.accounts import Accounts
from tikka.domains.entities.category import Category
from tikka.domains.entities.events import CategoryEvent
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.repository.categories import (
    CategoriesRepositoryInterface,
)


class Categories:

    """
    Categories domain class
    """

    def __init__(
        self,
        repository: CategoriesRepositoryInterface,
        accounts: Accounts,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Categories domain

        :param repository: AccountsRepositoryInterface instance
        :param accounts: Accounts domain instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.accounts = accounts
        self.event_dispatcher = event_dispatcher

    @staticmethod
    def create(name: str, parent_id: Optional[UUID] = None):
        """
        Return a new Category instance with a unique ID

        :param name: Name of the category
        :param parent_id: UUID of the parent category
        :return:
        """
        return Category(id=uuid4(), name=name, parent_id=parent_id)

    def add(self, category: Category):
        """
        Add category in repository

        :param category: Category instance
        :return:
        """
        self.repository.add(category)
        self.event_dispatcher.dispatch_event(
            CategoryEvent(CategoryEvent.EVENT_TYPE_ADD, category)
        )

    def update(self, category: Category):
        """
        Update category in repository

        :param category: Category instance
        :return:
        """
        self.repository.update(category)
        self.event_dispatcher.dispatch_event(
            CategoryEvent(CategoryEvent.EVENT_TYPE_UPDATE, category)
        )

    def get(self, id: UUID) -> Optional[Category]:  # pylint: disable=redefined-builtin
        """
        Get category instance

        :param id: Category ID
        :return:
        """
        return self.repository.get(id)

    def delete(self, id: UUID) -> None:  # pylint: disable=redefined-builtin
        """
        Delete category in repository

        :param id: Category ID to delete
        :return:
        """
        category = self.get(id)
        if category is None:
            return None

        # move all child categories into parent category
        self.repository.update_parent_id(id, category.parent_id)

        # move all child accounts into parent category
        self.accounts.update_category_id(id, category.parent_id)

        self.repository.delete(id)
        self.event_dispatcher.dispatch_event(
            CategoryEvent(CategoryEvent.EVENT_TYPE_DELETE, category)
        )

        return None

    def list(self, parent_id: Optional[UUID]) -> List[Category]:
        """
        Return list of Category instances with parent_id

        :param parent_id: Parent ID
        :return:
        """
        return self.repository.list(parent_id)

    def list_all(self) -> List[Category]:
        """
        Return list with all Category instances

        :return:
        """
        return self.repository.list_all()

    def expand(self, category: Category):
        """
        Update category with expanded attribute to True

        :param category: Category instance
        :return:
        """
        category.expanded = True
        self.repository.update(category)

    def collapse(self, category: Category):
        """
        Update category with expanded attribute to False

        :param category: Category instance
        :return:
        """
        category.expanded = False
        self.repository.update(category)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from queue import Queue
from threading import Thread

from substrateinterface import SubstrateInterface


class ThreadSafe(Thread):

    queue: Queue = Queue()

    def __init__(self, *args, **kwargs):
        """
        Init a SubstrateInterface client adapter instance as a thread

        :param args: Positional arguments
        :param kwargs: Keywords arguments
        """
        super().__init__(*args, **kwargs)

    def run(self):
        """
        Started asynchronously with Thread.start()

        :return:
        """
        while True:
            # print("loop...")
            call, method, args, result_handler, result = self.queue.get()
            result_ = dict()
            # print(call, method, args, result_handler, result)
            if call == "--close--":
                logging.debug("Close queue thread on substrate_interface")
                break

            try:
                # logging.debug(f"threadsafe call to rpc method {method}")
                result_ = call(method, args, result_handler)
            except Exception as exception:
                logging.error(method)
                logging.error(args)
                # logging.exception(exception)
                result.put(exception)
            # print(call.__name__, " put result ", result_)
            result.put(result_)
            # print("reloop...")

        logging.debug("SubstrateInterface connection closed and thread terminated.")

    def close(self):
        """
        Close connection

        :return:
        """
        # Closing the connection
        self.queue.put(("--close--", None, None, None, None))


class ThreadSafeSubstrateInterface(SubstrateInterface):
    """
    Override substrate_interface client class with a queue to be thread safe

    """

    def __init__(self, *args, **kwargs):
        """
        Init a SubstrateInterface client adapter instance as a thread

        :param args: Positional arguments
        :param kwargs: Keywords arguments
        """
        # create and start thread before calling parent init (which makes a rpc_request!)
        self.thread = ThreadSafe()
        self.thread.start()

        super().__init__(*args, **kwargs)

    def rpc_request(self, method, params, result_handler=None) -> dict:
        """
        Override rpc_request method to use threadsafe queue

        :param method: Name of the RPC method
        :param params: Params of the RPC method
        :param result_handler: Optional variable to receive results, default to None
        :return:
        """
        # print("queued rpc_request call !!")
        result: Queue = Queue()
        self.thread.queue.put(
            (super().rpc_request, method, params, result_handler, result)
        )
        # print(self.thread.queue.get())
        # print('done calling %s' % method)
        return_ = result.get()
        if isinstance(return_, Exception):
            raise return_
        return return_

    def close(self):
        logging.debug("Close RPC connection thread")
        self.thread.close()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Dewif lib module
"""
import base64
import hashlib
import random
import struct
from dataclasses import dataclass

from duniterpy.key.scrypt_params import ScryptParams
from mnemonic import Mnemonic

DEWIF_CURRENCY_CODE_NONE = 0x00000000
DEWIF_CURRENCY_CODE_G1 = 0x00000001
DEWIF_CURRENCY_CODE_G1_TEST = 0x10000001
DEWIF_ALGORITHM_CODE_ED25519 = 0x00
DEWIF_ALGORITHM_CODE_BIP32_ED25519 = 0x01
DEWIF_MNEMONIC_LANGUAGES = {
    0: "english",
    1: "chinese_simplified",
    2: "chinese_traditional",
    3: "french",
    4: "italian",
    5: "japanese",
    6: "korean",
    7: "spanish",
}


def random_bytes(size: int) -> bytes:
    """
    Return random bytes of given length

    :param size: Size of nonce in bytes
    :return:
    """
    return bytearray(random.getrandbits(8) for _ in range(size))


def byte_xor(bytes1, bytes2):
    """
    XOR two byte strings

    :param bytes1: First string
    :param bytes2: Second string
    :return:
    """
    return bytes([_a ^ _b for _a, _b in zip(bytes1, bytes2)])


def scrypt_maxmem(n_value, r_value, p_value):
    """
    Get sufficient maxmem scrypt parameter

    :param n_value: N value
    :param r_value: r value
    :param p_value: p value
    :return:
    """
    return 128 * r_value * (n_value + p_value + 2)


@dataclass
class DewifV1:
    """
    Dewif master class
    """

    password: str
    currency: int
    log_n: int

    version = 1

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def __str__(self) -> str:
        return base64.b64encode(bytes(self)).decode("utf-8")

    def __repr__(self):
        return f"class Ed25519 ({self.__str__()})"


@dataclass
class Ed25519V1(DewifV1):
    """
    Ed25519 algorithm class
    """

    seed: bytes
    pubkey: bytes

    algorithm = DEWIF_ALGORITHM_CODE_ED25519

    def __bytes__(self):
        nonce = random_bytes(12)

        # # test RFC
        # nonce = 0x013194f1286512cf094295cb.to_bytes(12, "big")

        header = struct.pack(">ii", self.version, self.currency)
        data = (
            header
            + self.log_n.to_bytes(1, "big")
            + self.algorithm.to_bytes(1, "big")
            + nonce
        )

        scrypt_params = ScryptParams()
        n_param = pow(2, self.log_n)
        salt = b"dewif" + nonce + self.password.encode("utf-8")
        xor_key = hashlib.scrypt(
            password=self.password.encode("utf-8"),
            salt=hashlib.sha256(salt).digest(),
            n=n_param,
            r=scrypt_params.r,  # 16
            p=scrypt_params.p,  # 1
            dklen=64,
            maxmem=scrypt_maxmem(n_param, scrypt_params.r, scrypt_params.p),
        )
        data_to_encrypt = self.seed + self.pubkey
        encrypted_data = byte_xor(data_to_encrypt, xor_key)
        data += encrypted_data

        # test RFC
        # assert base64.b64encode(data).decode("utf-8") == \
        # "AAAAARAAAAEMAAExlPEoZRLPCUKVy0iKnn1HUSFcmhwJPQETAghFDvH8ZmX59IuvR9hYV1gnVjCpU+TGOdUzyQmj3+auw3vUpFQYBiRlh67/I1xAhZM="

        return data


@dataclass
class Bip32Ed25519V1(DewifV1):
    """
    BIP32-Ed25519 algorithm class
    """

    mnemonic: str
    language: str

    algorithm = DEWIF_ALGORITHM_CODE_BIP32_ED25519

    def __bytes__(self):
        nonce = random_bytes(12)

        # test RFC
        # nonce = 0xc54299ae71fe2a4ecdc7d58a.to_bytes(12, "big")

        header = struct.pack(">ii", self.version, self.currency)
        data = (
            header
            + self.log_n.to_bytes(1, "big")
            + self.algorithm.to_bytes(1, "big")
            + nonce
        )

        scrypt_params = ScryptParams()
        n_param = pow(2, self.log_n)
        salt = b"dewif" + nonce + self.password.encode("utf-8")
        xor_key = hashlib.scrypt(
            password=self.password.encode("utf-8"),
            salt=hashlib.sha256(salt).digest(),
            n=pow(2, self.log_n),
            r=scrypt_params.r,  # 16
            p=scrypt_params.p,  # 1
            dklen=42,
            maxmem=scrypt_maxmem(n_param, scrypt_params.r, scrypt_params.p),
        )

        entropy = Mnemonic(self.language).to_entropy(self.mnemonic)
        entropy_length = len(entropy)
        entropy_padding = random_bytes(32 - entropy_length)

        # test RFC
        # entropy_padding = 0xaa083bd16c8317121d34b5aed1c1420a.to_bytes(16, "big")

        language_code = list(DEWIF_MNEMONIC_LANGUAGES.keys())[
            list(DEWIF_MNEMONIC_LANGUAGES.values()).index(self.language)
        ]
        checksum = hashlib.sha256(
            nonce
            + language_code.to_bytes(1, "big")
            + entropy_length.to_bytes(1, "big")
            + entropy
        ).digest()[:8]

        data_to_encrypt = (
            language_code.to_bytes(1, "big")
            + entropy_length.to_bytes(1, "big")
            + entropy
            + entropy_padding
            + checksum
        )
        encrypted_data = byte_xor(data_to_encrypt, xor_key)
        data += encrypted_data

        # test RFC
        # assert base64.b64encode(data).decode("utf-8") == \
        # "AAAAARAAAAEOAcVCma5x/ipOzcfViufNdfj5k4Sl5zdrHLf9PPGDkH1Pz3y8tFrx/jZZcJd92LIk+EWIrjxiSw=="

        return data



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from tikka.domains.entities.identity import Identity


class IdentitiesRepositoryInterface(abc.ABC):
    """
    IdentitiesRepositoryInterface class
    """

    COLUMN_INDEX = "index_"

    @abc.abstractmethod
    def add(self, identity: Identity) -> None:
        """
        Add a new identity in repository

        :param identity: Identity instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, index: int) -> Optional[Identity]:
        """
        Return Identity instance from repository

        :param index: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_address(self, address: str) -> Optional[Identity]:
        """
        Return Identity instance from repository

        :param address: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_index_by_address(self, address: str) -> Optional[int]:
        """
        Return Identity index from repository

        :param address: Identity index
        :return:
        """
        raise NotImplementedError

    def update(self, identity: Identity) -> None:
        """
        Update identity in repository

        :param identity: Identity instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, index: int) -> None:
        """
        Delete identity in repository

        :param index: Identity index to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, index: int) -> bool:
        """
        Return True if identity with index is in repository, else False

        :param index: Identity index to check
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any, List, Optional

from PyQt5.QtCore import QAbstractItemModel, QLocale, QModelIndex, Qt, QVariant
from PyQt5.QtGui import QFont, QIcon, QPixmap

from tikka.domains.application import Application
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY
from tikka.domains.entities.identity import IdentityStatus
from tikka.domains.entities.smith import SmithStatus
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.interfaces.adapters.repository.identities import (
    IdentitiesRepositoryInterface,
)
from tikka.slots.pyqt.entities.constants import (
    ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY,
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_IDENTITY,
    ICON_SMITH,
    SELECTED_UNIT_PREFERENCES_KEY,
)


class AccountTableModel(QAbstractItemModel):
    """
    AccountTableModel class that drives the population of table display
    """

    REPOSITORY_COLUMNS = [
        "",
        IdentitiesRepositoryInterface.COLUMN_INDEX,
        AccountsRepositoryInterface.COLUMN_BALANCE,
        AccountsRepositoryInterface.COLUMN_NAME,
        AccountsRepositoryInterface.COLUMN_ADDRESS,
        AccountsRepositoryInterface.COLUMN_PATH,
        AccountsRepositoryInterface.COLUMN_ROOT,
        AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE,
        AccountsRepositoryInterface.COLUMN_CATEGORY_ID,
    ]

    def __init__(self, application: Application, locale: QLocale):
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext
        # get parent widget locale to localize balance display
        self.locale = locale

        # calculate balance display by unit preference
        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # drag and drop mime-type
        self.mime_type = "application/vnd.text.list"

        # Number of column displayed, see self.data()
        self._column_count = len(AccountTableModel.REPOSITORY_COLUMNS)

        self.table_view_data: List[AccountsRepositoryInterface.TableViewRow] = []

        self.init_data()

    def init_data(self):
        """
        Fill data from repository

        :return:
        """
        # calculate balance display by unit preference
        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        sort_column_index = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if sort_column_index is not None:
            sort_column_index = int(sort_column_index)
        repository_sort_order = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )

        filters = {}
        category_id_filter = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY
        )
        if category_id_filter is not None:
            filters[
                AccountsRepositoryInterface.TABLE_VIEW_FILTER_BY_CATEGORY_ID
            ] = category_id_filter

        wallet_filter_preference = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY
        )
        if wallet_filter_preference is not None:
            # preference store boolean as integer in a string column ("0" or "1")
            # convert it to boolean
            wallet_filter = int(wallet_filter_preference) == 1
            filters[
                AccountsRepositoryInterface.TABLE_VIEW_FILTER_BY_WALLET
            ] = wallet_filter

        if len(filters) == 0:
            filters = None

        self.beginResetModel()
        self.table_view_data = self.application.accounts.repository.table_view(
            filters,
            sort_column_index=sort_column_index,
            sort_order=repository_sort_order,
        )
        self.endResetModel()

    def rowCount(
        self, parent: QModelIndex = QModelIndex()  # pylint: disable=unused-argument
    ) -> int:
        """
        Return row count from account list

        :param parent: QModelIndex instance
        :return:
        """
        return len(self.table_view_data)

    def columnCount(
        self, parent: QModelIndex = QModelIndex()  # pylint: disable=unused-argument
    ) -> int:
        """
        Return column count of parent QModelIndex

        :param parent: QModelIndex instance
        :return:
        """
        return self._column_count

    def index(
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        """
        Return QModelIndex for row, column and parent

        :param row: Row index
        :param column: Column index
        :param parent: Parent QModelIndex instance
        :return:
        """
        if not QAbstractItemModel.hasIndex(self, row, column, parent):
            return QModelIndex()

        if row < len(self.table_view_data):
            return QAbstractItemModel.createIndex(
                self, row, column, self.table_view_data[row]
            )

        return QModelIndex()

    def parent(
        self, child: Optional[QModelIndex] = None
    ):  # pylint: disable=unused-argument
        """
        Return parent QModelIndex of child QModelIndex

        :param child: QModelIndex instance
        :return:
        """
        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        data = QVariant()
        if not index.isValid():
            return data

        table_view_row = self.table_view_data[index.row()]

        if role == Qt.FontRole:
            font = QFont()
            # display addresses in monospace font
            if index.column() == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ADDRESS
            ) or index.column() == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ROOT
            ):
                font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
                # if font is not found, choose equivalent monospace font on system
                font.setStyleHint(QFont.Monospace)

            # display root accounts with underline
            if table_view_row.root is None:
                font.setUnderline(True)

            return QVariant(font)

        # align right for balance column
        if (
            role == Qt.TextAlignmentRole
            and index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_BALANCE
            )
        ):
            return Qt.AlignRight

        # display account properties
        # display wallet lock status icon for account in first column
        if (
            index.column() == AccountTableModel.REPOSITORY_COLUMNS.index("")
            and role == Qt.DecorationRole
        ):
            if (
                table_view_row.wallet_address is not None
                and self.application.wallets.exists(table_view_row.wallet_address)
            ):
                data = (
                    QVariant(QIcon(QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED)))
                    if self.application.wallets.is_unlocked(
                        table_view_row.wallet_address
                    )
                    else QVariant(QIcon(QPixmap(ICON_ACCOUNT_WALLET_LOCKED)))
                )
            else:
                data = QVariant(QIcon(QPixmap(ICON_ACCOUNT_NO_WALLET)))
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                IdentitiesRepositoryInterface.COLUMN_INDEX
            )
            and role == Qt.DecorationRole
        ):
            identity = self.application.identities.get_by_address(
                table_view_row.address
            )
            if identity is not None and identity.status == IdentityStatus.MEMBER:
                smith = self.application.smiths.get(identity.index)
                if smith is not None and smith.status == SmithStatus.SMITH:
                    data = QVariant(
                        QPixmap(ICON_SMITH).scaled(
                            16, 16, aspectRatioMode=Qt.KeepAspectRatio
                        )
                    )
                else:
                    data = QVariant(
                        QPixmap(ICON_IDENTITY).scaled(
                            16, 16, aspectRatioMode=Qt.KeepAspectRatio
                        )
                    )
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_BALANCE
            )
            and role == Qt.DisplayRole
            and table_view_row.balance is not None
        ):
            data = QVariant(
                self.locale.toString(self.amount.value(table_view_row.balance), "f", 2)
            )
        elif index.column() == AccountTableModel.REPOSITORY_COLUMNS.index(
            AccountsRepositoryInterface.COLUMN_NAME
        ) and role in (Qt.DisplayRole, Qt.EditRole):
            data = QVariant(table_view_row.name)
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ADDRESS
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(table_view_row.address)
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_PATH
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(table_view_row.path)
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ROOT
            )
            and role == Qt.DisplayRole
            and table_view_row.root is not None
        ):
            data = QVariant(table_view_row.root)
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE
            )
            and role == Qt.DisplayRole
        ):
            data = (
                QVariant("SR25519")
                if table_view_row.crypto_type == 1
                else QVariant("ED25519")
            )
        elif (
            index.column()
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_CATEGORY_ID
            )
            and role == Qt.DisplayRole
        ):
            data = (
                QVariant(table_view_row.category_name)
                if table_view_row.category_name is not None
                else QVariant()
            )

        return data

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,  # pylint: disable=unused-argument
        role: int = Qt.DisplayRole,
    ) -> QVariant:
        """
        Return header data

        :param section: Section offset
        :param orientation: Qt Orientation flag
        :param role: Qt Role flag
        :return:
        """
        data = QVariant()
        if (
            section == AccountTableModel.REPOSITORY_COLUMNS.index("")
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Wallet"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                IdentitiesRepositoryInterface.COLUMN_INDEX
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Identity"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_BALANCE
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(
                self._("Balance ({symbol})").format(symbol=self.amount.symbol())
            )
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_NAME
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Name"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ADDRESS
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Address"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_PATH
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Derivation"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_ROOT
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Root"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Crypto"))
        elif (
            section
            == AccountTableModel.REPOSITORY_COLUMNS.index(
                AccountsRepositoryInterface.COLUMN_CATEGORY_ID
            )
            and role == Qt.DisplayRole
        ):
            data = QVariant(self._("Category"))

        return data

    def sort(self, column: int, order: Qt.SortOrder = None) -> None:
        """
        Sort by column number in order

        :param column: Column index
        :param order: Qt.SortOrder flag
        :return:
        """
        if column > -1:
            sort_order = (
                AccountsRepositoryInterface.SORT_ORDER_ASCENDING
                if order == Qt.AscendingOrder
                else AccountsRepositoryInterface.SORT_ORDER_DESCENDING
            )
            self.application.preferences_repository.set(
                ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY, str(column)
            )
            self.application.preferences_repository.set(
                ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY, sort_order
            )
            self.init_data()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tikka.interfaces.entities.amount import AmountInterface


class UnitAmount(AmountInterface):
    """
    UnitAmount class
    """

    def value(self, blockchain_value: int) -> float:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.value.__doc__
        )
        decimals = (
            2
            if self.currencies.get_current().token_decimals is None
            else self.currencies.get_current().token_decimals
        )
        return blockchain_value / pow(10, decimals)  # type: ignore

    def blockchain_value(self, value: float) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.blockchain_value.__doc__
        )
        decimals = (
            2
            if self.currencies.get_current().token_decimals is None
            else self.currencies.get_current().token_decimals
        )
        return round(value * pow(10, decimals))  # type: ignore

    def symbol(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.symbol.__doc__
        )
        symbol = (
            self.currencies.get_current().name
            if self.currencies.get_current().token_symbol is None
            else self.currencies.get_current().token_symbol
        )
        return symbol  # type: ignore

    def name(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.name.__doc__
        )
        return self._("Units")


class UniversalDividendAmount(AmountInterface):
    """
    UniversalDividendAmount class
    """

    def value(self, blockchain_value: int) -> float:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.value.__doc__
        )
        if self.currencies.get_current().universal_dividend is None:
            return 0
        return blockchain_value / self.currencies.get_current().universal_dividend  # type: ignore

    def blockchain_value(self, value: float) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.blockchain_value.__doc__
        )
        if self.currencies.get_current().universal_dividend is None:
            return 0
        return round(value * self.currencies.get_current().universal_dividend)  # type: ignore

    def symbol(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.symbol.__doc__
        )
        symbol = (
            self.currencies.get_current().name
            if self.currencies.get_current().token_symbol is None
            else self.currencies.get_current().token_symbol
        )
        return self._("UD {symbol}").format(symbol=symbol)

    def name(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AmountInterface.name.__doc__
        )
        return self._("Universal Dividend")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/account.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountWidget(object):
    def setupUi(self, AccountWidget):
        AccountWidget.setObjectName("AccountWidget")
        AccountWidget.resize(615, 274)
        AccountWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(AccountWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.QRCodeAddressLabel = QtWidgets.QLabel(AccountWidget)
        self.QRCodeAddressLabel.setMaximumSize(QtCore.QSize(100, 100))
        self.QRCodeAddressLabel.setBaseSize(QtCore.QSize(200, 0))
        self.QRCodeAddressLabel.setText("")
        self.QRCodeAddressLabel.setPixmap(QtGui.QPixmap(":/images/safe"))
        self.QRCodeAddressLabel.setScaledContents(True)
        self.QRCodeAddressLabel.setObjectName("QRCodeAddressLabel")
        self.horizontalLayout.addWidget(self.QRCodeAddressLabel)
        self.balanceLabel = QtWidgets.QLabel(AccountWidget)
        font = QtGui.QFont()
        font.setPointSize(40)
        self.balanceLabel.setFont(font)
        self.balanceLabel.setObjectName("balanceLabel")
        self.horizontalLayout.addWidget(self.balanceLabel)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.refreshButton = QtWidgets.QPushButton(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshButton.setIcon(icon)
        self.refreshButton.setObjectName("refreshButton")
        self.verticalLayout.addWidget(self.refreshButton, 0, QtCore.Qt.AlignLeft)
        self.transferToButton = QtWidgets.QPushButton(AccountWidget)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/money-recive-svgrepo-com.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.transferToButton.setIcon(icon1)
        self.transferToButton.setIconSize(QtCore.QSize(32, 32))
        self.transferToButton.setObjectName("transferToButton")
        self.verticalLayout.addWidget(self.transferToButton)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setContentsMargins(6, -1, 0, -1)
        self.formLayout_2.setObjectName("formLayout_2")
        self.lockStatusIcon = QtWidgets.QLabel(AccountWidget)
        self.lockStatusIcon.setMaximumSize(QtCore.QSize(16, 16))
        self.lockStatusIcon.setText("")
        self.lockStatusIcon.setPixmap(QtGui.QPixmap(":/images/locked_cartoon_400x400.png"))
        self.lockStatusIcon.setScaledContents(True)
        self.lockStatusIcon.setObjectName("lockStatusIcon")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lockStatusIcon)
        self.addressValueLabel = QtWidgets.QLabel(AccountWidget)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.addressValueLabel)
        self.identityLabel = QtWidgets.QLabel(AccountWidget)
        self.identityLabel.setMaximumSize(QtCore.QSize(16, 20))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.identityLabel.setFont(font)
        self.identityLabel.setText("")
        self.identityLabel.setPixmap(QtGui.QPixmap(":/icons/member_icon"))
        self.identityLabel.setScaledContents(True)
        self.identityLabel.setObjectName("identityLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.identityLabel)
        self.identityValueLabel = QtWidgets.QLabel(AccountWidget)
        self.identityValueLabel.setObjectName("identityValueLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.identityValueLabel)
        self.nameLabel = QtWidgets.QLabel(AccountWidget)
        self.nameLabel.setText("")
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.nameLabel)
        self.derivationIconLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.derivationIconLabel.sizePolicy().hasHeightForWidth())
        self.derivationIconLabel.setSizePolicy(sizePolicy)
        self.derivationIconLabel.setMaximumSize(QtCore.QSize(16, 20))
        self.derivationIconLabel.setText("")
        self.derivationIconLabel.setPixmap(QtGui.QPixmap(":/icons/wot_icon"))
        self.derivationIconLabel.setScaledContents(True)
        self.derivationIconLabel.setObjectName("derivationIconLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.derivationIconLabel)
        self.derivationValueLabel = QtWidgets.QLabel(AccountWidget)
        self.derivationValueLabel.setText("")
        self.derivationValueLabel.setObjectName("derivationValueLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.derivationValueLabel)
        self.smithLabel = QtWidgets.QLabel(AccountWidget)
        self.smithLabel.setMaximumSize(QtCore.QSize(16, 20))
        self.smithLabel.setText("")
        self.smithLabel.setPixmap(QtGui.QPixmap(":/icons/smith"))
        self.smithLabel.setScaledContents(True)
        self.smithLabel.setObjectName("smithLabel")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.smithLabel)
        self.smithValuelabel = QtWidgets.QLabel(AccountWidget)
        self.smithValuelabel.setObjectName("smithValuelabel")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.smithValuelabel)
        self.verticalLayout_2.addLayout(self.formLayout_2)
        self.errorLabel = QtWidgets.QLabel(AccountWidget)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout_2.addWidget(self.errorLabel)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)

        self.retranslateUi(AccountWidget)
        QtCore.QMetaObject.connectSlotsByName(AccountWidget)

    def retranslateUi(self, AccountWidget):
        #  _translate = QtCore.QCoreApplication.translate
        AccountWidget.setWindowTitle(_("Account"))
        self.balanceLabel.setText(_("0"))
        self.refreshButton.setText(_("Refresh"))
        self.transferToButton.setText(_("Transfer to"))
        self.identityValueLabel.setText(_("None"))
        self.smithValuelabel.setText(_("None"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from enum import Enum
from typing import Optional


class IdentityStatus(int, Enum):
    UNCONFIRMED = 0
    UNVALIDATED = 1
    MEMBER = 2
    NOT_MEMBER = 3
    REVOKED = 4


@dataclass
class Identity:
    index: int
    removable_on: int
    next_creatable_on: int
    status: IdentityStatus
    address: str
    old_address: Optional[str]


@dataclass
class Certification:
    issuer_identity_index: int
    issuer_address: str
    receiver_identity_index: int
    receiver_address: str
    expire_on_block: int



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List, Optional

from tikka.domains.entities.node import Node


class NodesRepositoryInterface(abc.ABC):
    """
    NodesRepositoryInterface class
    """

    COLUMN_URL = "url"
    COLUMN_PEER_ID = "peer_id"
    COLUMN_BLOCK = "block"
    COLUMN_SOFTWARE = "software"
    COLUMN_SOFTWARE_VERSION = "software_version"
    COLUMN_SESSION_KEYS = "session_keys"
    COLUMN_EPOCH_INDEX = "epoch_index"
    COLUMN_UNSAFE_API_EXPOSED = "unsafe_api_exposed"

    DEFAULT_LIST_OFFSET = 0
    DEFAULT_LIST_LIMIT = 1000

    @abc.abstractmethod
    def add(self, node: Node) -> None:
        """
        Add a new node in repository

        :param node: Node instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, url: str) -> Optional[Node]:
        """
        Return Node by url from repository

        :param url: Node url
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, node: Node) -> None:
        """
        Update node in repository

        :param node: Node instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(
        self,
        offset: int = DEFAULT_LIST_OFFSET,
        limit: int = DEFAULT_LIST_LIMIT,
    ) -> List[Node]:
        """
        List nodes from repository

        :param offset: Offset index to get rows from
        :param limit: Number of rows to return
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, url: str) -> None:
        """
        Delete node by url in repository

        :param url: Url of Node to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all nodes in repository

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def count(self) -> int:
        """
        Return number of nodes in repository

        :return:
        """
        raise NotImplementedError

    def get_urls(self) -> List[str]:
        """
        Get all urls of nodes from repository

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
from datetime import datetime
from typing import Any, List, Optional

from tikka.adapters.repository.sqlite3 import Sqlite3RepositoryInterface
from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.interfaces.adapters.repository.smiths import SmithsRepositoryInterface

TABLE_NAME = "smiths"


class Sqlite3SmithsRepository(SmithsRepositoryInterface, Sqlite3RepositoryInterface):
    """
    Sqlite3SmithsRepository class
    """

    def add(self, smith: Smith) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_smith(smith),
        )

    def get(self, index: int) -> Optional[Smith]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.get.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE identity_index=?", (index,)
        )
        if row is None:
            return None

        return get_smith_from_row(row)

    def update(self, smith: Smith) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"identity_index='{smith.identity_index}'",
            **get_fields_from_smith(smith),
        )

    def delete(self, index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, identity_index=index)

    def exists(self, index: int) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.exists.__doc__
        )

        row = self.client.select_one(
            f"SELECT count(identity_index) FROM {TABLE_NAME} WHERE identity_index=?",
            (index,),
        )
        if row is None:
            return False

        return row[0] == 1

    def list(self) -> List[Smith]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            SmithsRepositoryInterface.list.__doc__
        )
        result_set = self.client.select(
            f"SELECT * FROM {TABLE_NAME} ORDER BY identity_index ASC"
        )

        list_ = []
        for row in result_set:
            list_.append(get_smith_from_row(row))

        return list_


def get_fields_from_smith(smith: Smith) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param smith: Smith instance
    :return:
    """
    fields = {}
    for (key, value) in smith.__dict__.items():
        if key.startswith("_"):
            continue
        if key == "status":
            # convert SmithStatus Enum to int
            value = value.value
        if key == "expire_on" and value is not None:
            # convert datetime to int
            value = int(value.timestamp())
        if key == "certifications_received" or key == "certifications_issued":
            value = json.dumps(value)
        fields[key] = value

    return fields


def get_smith_from_row(row: tuple) -> Smith:
    """
    Return a Smith instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 1:
            # convert int to SmithStatus Enum
            values.append(SmithStatus(value))
        elif count == 2 and value is not None:
            # convert timestamp to datetime
            values.append(datetime.fromtimestamp(value))
        elif count == 3 or count == 4:
            # convert certifications json string to list
            values.append(json.loads(value))
        else:
            values.append(value)
        count += 1

    return Smith(*values)  # pylint: disable=no-value-for-parameter



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import random
from typing import List, Optional

from tikka.adapters.network.nodes import NetworkNodes
from tikka.adapters.network.rpc.connection import RPCConnection
from tikka.domains.config import Config
from tikka.domains.connections import Connections
from tikka.domains.currencies import Currencies
from tikka.domains.entities.constants import (
    NODES_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
)
from tikka.domains.entities.events import ConnectionsEvent, NodesEvent
from tikka.domains.entities.node import Node
from tikka.domains.events import EventDispatcher
from tikka.domains.preferences import Preferences
from tikka.interfaces.adapters.network.nodes import NetworkNodesInterface
from tikka.interfaces.adapters.repository.nodes import NodesRepositoryInterface


class Nodes:
    """
    Nodes domain class
    """

    def __init__(
        self,
        repository: NodesRepositoryInterface,
        preferences: Preferences,
        connections: Connections,
        network: NetworkNodesInterface,
        config: Config,
        currencies: Currencies,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Nodes domain instance

        :param repository: EntryPointsRepositoryInterface instance
        :param preferences: Preferences domain instance
        :param connections: Connections instance
        :param network: Network adapter instance for handling nodes
        :param config: Config instance
        :param currencies: Currencies instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.preferences = preferences
        self.connections = connections
        self.network = network
        self.config = config
        self.currencies = currencies
        self.event_dispatcher = event_dispatcher
        self._current_url = self.currencies.get_entry_point_urls()[0]

        self.init_repository()

        # events
        self.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_CONNECTED, self._on_connections_connected
        )

    def init_repository(self):
        """
        Init repository with default entry points from config

        :return:
        """
        repository_urls = self.repository.get_urls()

        # init repository with current currency entry point urls
        for url in self.currencies.get_entry_point_urls():
            if url not in repository_urls:
                self.repository.add(Node(url))

        self._current_url = self.repository.list(0, 1)[0].url

        current_url_in_preferences = self.preferences.get(
            NODES_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY
        )
        if (
            current_url_in_preferences is None
            or current_url_in_preferences not in self.repository.get_urls()
        ):
            self.preferences.set(
                NODES_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
            )
        else:
            self._current_url = current_url_in_preferences

    def network_fetch_current_node(self) -> None:
        """
        Update node from network

        :return:
        """
        current_node = self.repository.get(self.get_current_url())
        network_node = self.network.get()
        if network_node is None:
            return None

        if current_node is not None:
            # update only changing properties
            current_node.block = network_node.block
            current_node.peer_id = network_node.peer_id
            current_node.software = network_node.software
            current_node.software_version = network_node.software_version
            current_node.epoch_index = network_node.epoch_index
            current_node.unsafe_api_exposed = network_node.unsafe_api_exposed

            self.repository.update(current_node)

        return None

    @staticmethod
    def network_test_and_get_node(url: str) -> Optional[Node]:
        """
        Try to open connection on url and return node if successful

        Then close connection

        :param url: Entry point url
        :return:
        """
        node = None

        connections = Connections(RPCConnection(), EventDispatcher())
        connections.connect(Node(url))
        if connections.is_connected():
            network_nodes = NetworkNodes(connections)
            node = network_nodes.get()
        connections.disconnect()

        return node

    def set_current_url_randomly(self):
        """
        Shuffle node list randomly and connect to first available node

        :return:
        """
        indices = list(range(1, self.count()))
        random.shuffle(indices)
        for index in indices:
            url = self.list()[index].url
            # never choose localhost randomly...
            if "localhost" not in url:
                self.set_current_url(self.list()[index].url)
                if self.connections.is_connected():
                    break

    def add(self, node: Node) -> None:
        """
        Add node in repository

        :param node: Node instance
        :return:
        """
        self.repository.add(node)

        self.event_dispatcher.dispatch_event(
            NodesEvent(NodesEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def get(self, url: str) -> Optional[Node]:
        """
        Get Node instance by url

        :param url: Url
        :return:
        """
        return self.repository.get(url)

    def update(self, node: Node) -> None:
        """
        Update Node in repository

        :param node: Node instance
        :return:
        """
        self.repository.update(node)

        self.event_dispatcher.dispatch_event(
            NodesEvent(NodesEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def list(self) -> List[Node]:
        """
        Return all Nodes from repository

        :return:
        """
        return self.repository.list()

    def count(self) -> int:
        """
        Return total node count

        :return:
        """
        return self.repository.count()

    def delete(self, url: str):
        """
        Delete Node by url

        :param url: Node url
        :return:
        """
        # do not delete default entry points from config
        if url in self.currencies.get_entry_point_urls():
            return

        self.repository.delete(url)
        # switch current entry point to first in list
        self.set_current_url(self.repository.list(0, 1)[0].url)
        # set new entry point in preferences
        self.preferences.set(
            NODES_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
        )

        self.event_dispatcher.dispatch_event(
            NodesEvent(NodesEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def get_current_url(self) -> str:
        """
        Return current entry point url

        :return:
        """
        return self._current_url

    def set_current_url(self, url: str) -> None:
        """
        Set current entry point url

        :return:
        """
        self._current_url = url
        # update preference
        self.preferences.set(
            NODES_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
            self._current_url,
        )
        # switch current connection
        self.connections.disconnect()
        node = self.get(self._current_url)
        if node is not None:
            self.connections.connect(node)

    def _on_connections_connected(self, _: ConnectionsEvent):
        """
        Triggered when the connection is established

        :param _:
        :return:
        """
        self.network_fetch_current_node()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/v1_account_import_wizard.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_importAccountV1Wizard(object):
    def setupUi(self, importAccountV1Wizard):
        importAccountV1Wizard.setObjectName("importAccountV1Wizard")
        importAccountV1Wizard.resize(878, 421)
        importAccountV1Wizard.setModal(True)
        self.wizardPage1 = QtWidgets.QWizardPage()
        self.wizardPage1.setObjectName("wizardPage1")
        self.page1_label = QtWidgets.QLabel(self.wizardPage1)
        self.page1_label.setGeometry(QtCore.QRect(0, 0, 741, 261))
        self.page1_label.setWordWrap(True)
        self.page1_label.setObjectName("page1_label")
        importAccountV1Wizard.addPage(self.wizardPage1)
        self.wizardPage2 = QtWidgets.QWizardPage()
        self.wizardPage2.setObjectName("wizardPage2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.wizardPage2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.page2_label = QtWidgets.QLabel(self.wizardPage2)
        font = QtGui.QFont()
        font.setItalic(False)
        self.page2_label.setFont(font)
        self.page2_label.setAlignment(QtCore.Qt.AlignCenter)
        self.page2_label.setWordWrap(True)
        self.page2_label.setObjectName("page2_label")
        self.verticalLayout.addWidget(self.page2_label)
        self.groupBox = QtWidgets.QGroupBox(self.wizardPage2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.groupBox.setFont(font)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setContentsMargins(0, 0, -1, 0)
        self.gridLayout.setHorizontalSpacing(6)
        self.gridLayout.setObjectName("gridLayout")
        self.sourceAddressValueLabel = QtWidgets.QLabel(self.groupBox)
        self.sourceAddressValueLabel.setText("")
        self.sourceAddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.sourceAddressValueLabel.setObjectName("sourceAddressValueLabel")
        self.gridLayout.addWidget(self.sourceAddressValueLabel, 3, 2, 1, 1)
        self.sourceV1AddressLabel = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceV1AddressLabel.sizePolicy().hasHeightForWidth())
        self.sourceV1AddressLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceV1AddressLabel.setFont(font)
        self.sourceV1AddressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceV1AddressLabel.setObjectName("sourceV1AddressLabel")
        self.gridLayout.addWidget(self.sourceV1AddressLabel, 2, 0, 1, 1)
        self.sourceShowButton = QtWidgets.QPushButton(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceShowButton.sizePolicy().hasHeightForWidth())
        self.sourceShowButton.setSizePolicy(sizePolicy)
        self.sourceShowButton.setObjectName("sourceShowButton")
        self.gridLayout.addWidget(self.sourceShowButton, 0, 3, 1, 1, QtCore.Qt.AlignRight)
        self.sourceAddressLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceAddressLabel.setFont(font)
        self.sourceAddressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceAddressLabel.setObjectName("sourceAddressLabel")
        self.gridLayout.addWidget(self.sourceAddressLabel, 3, 0, 1, 1)
        self.sourcePasswordIDLineEdit = QtWidgets.QLineEdit(self.groupBox)
        self.sourcePasswordIDLineEdit.setEchoMode(QtWidgets.QLineEdit.Password)
        self.sourcePasswordIDLineEdit.setObjectName("sourcePasswordIDLineEdit")
        self.gridLayout.addWidget(self.sourcePasswordIDLineEdit, 1, 2, 1, 1)
        self.sourceSecretIDLineEdit = QtWidgets.QLineEdit(self.groupBox)
        self.sourceSecretIDLineEdit.setEchoMode(QtWidgets.QLineEdit.Password)
        self.sourceSecretIDLineEdit.setReadOnly(False)
        self.sourceSecretIDLineEdit.setObjectName("sourceSecretIDLineEdit")
        self.gridLayout.addWidget(self.sourceSecretIDLineEdit, 0, 2, 1, 1)
        self.sourcePasswordIDLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setItalic(False)
        font.setWeight(75)
        self.sourcePasswordIDLabel.setFont(font)
        self.sourcePasswordIDLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourcePasswordIDLabel.setObjectName("sourcePasswordIDLabel")
        self.gridLayout.addWidget(self.sourcePasswordIDLabel, 1, 0, 1, 1)
        self.sourceSecretIDLabel = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceSecretIDLabel.sizePolicy().hasHeightForWidth())
        self.sourceSecretIDLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceSecretIDLabel.setFont(font)
        self.sourceSecretIDLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceSecretIDLabel.setObjectName("sourceSecretIDLabel")
        self.gridLayout.addWidget(self.sourceSecretIDLabel, 0, 0, 1, 1)
        self.sourceBalanceValueLabel = QtWidgets.QLabel(self.groupBox)
        self.sourceBalanceValueLabel.setText("")
        self.sourceBalanceValueLabel.setObjectName("sourceBalanceValueLabel")
        self.gridLayout.addWidget(self.sourceBalanceValueLabel, 4, 2, 1, 1)
        self.sourceIdentityLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceIdentityLabel.setFont(font)
        self.sourceIdentityLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceIdentityLabel.setObjectName("sourceIdentityLabel")
        self.gridLayout.addWidget(self.sourceIdentityLabel, 5, 0, 1, 1)
        self.sourceBalanceLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceBalanceLabel.setFont(font)
        self.sourceBalanceLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceBalanceLabel.setObjectName("sourceBalanceLabel")
        self.gridLayout.addWidget(self.sourceBalanceLabel, 4, 0, 1, 1)
        self.sourceIdentityValueLabel = QtWidgets.QLabel(self.groupBox)
        self.sourceIdentityValueLabel.setText("")
        self.sourceIdentityValueLabel.setObjectName("sourceIdentityValueLabel")
        self.gridLayout.addWidget(self.sourceIdentityValueLabel, 5, 2, 1, 1)
        self.sourceV1AddressValueLabel = QtWidgets.QLabel(self.groupBox)
        self.sourceV1AddressValueLabel.setText("")
        self.sourceV1AddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.sourceV1AddressValueLabel.setObjectName("sourceV1AddressValueLabel")
        self.gridLayout.addWidget(self.sourceV1AddressValueLabel, 2, 2, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.sourceErrorLabel = QtWidgets.QLabel(self.groupBox)
        self.sourceErrorLabel.setText("")
        self.sourceErrorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.sourceErrorLabel.setObjectName("sourceErrorLabel")
        self.verticalLayout_2.addWidget(self.sourceErrorLabel)
        self.verticalLayout.addWidget(self.groupBox)
        importAccountV1Wizard.addPage(self.wizardPage2)
        self.wizardPage3 = QtWidgets.QWizardPage()
        self.wizardPage3.setObjectName("wizardPage3")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.wizardPage3)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.page3_label = QtWidgets.QLabel(self.wizardPage3)
        font = QtGui.QFont()
        font.setItalic(False)
        self.page3_label.setFont(font)
        self.page3_label.setAlignment(QtCore.Qt.AlignCenter)
        self.page3_label.setWordWrap(True)
        self.page3_label.setObjectName("page3_label")
        self.verticalLayout_4.addWidget(self.page3_label)
        self.groupBox_2 = QtWidgets.QGroupBox(self.wizardPage3)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.groupBox_2.setFont(font)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.GridLayout_2 = QtWidgets.QGridLayout()
        self.GridLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.GridLayout_2.setContentsMargins(-1, 0, -1, 0)
        self.GridLayout_2.setObjectName("GridLayout_2")
        self.destinationMnemonicLineEdit = QtWidgets.QLineEdit(self.groupBox_2)
        self.destinationMnemonicLineEdit.setEchoMode(QtWidgets.QLineEdit.Password)
        self.destinationMnemonicLineEdit.setObjectName("destinationMnemonicLineEdit")
        self.GridLayout_2.addWidget(self.destinationMnemonicLineEdit, 3, 1, 1, 1)
        self.destinationShowButton = QtWidgets.QPushButton(self.groupBox_2)
        self.destinationShowButton.setObjectName("destinationShowButton")
        self.GridLayout_2.addWidget(self.destinationShowButton, 3, 2, 1, 1)
        self.destinationAddressLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationAddressLabel.setFont(font)
        self.destinationAddressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationAddressLabel.setObjectName("destinationAddressLabel")
        self.GridLayout_2.addWidget(self.destinationAddressLabel, 6, 0, 1, 1)
        self.destinationRootNameOrAddressLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationRootNameOrAddressLabel.setFont(font)
        self.destinationRootNameOrAddressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationRootNameOrAddressLabel.setObjectName("destinationRootNameOrAddressLabel")
        self.GridLayout_2.addWidget(self.destinationRootNameOrAddressLabel, 1, 0, 1, 1)
        self.destinationRootAddressLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationRootAddressLabel.setFont(font)
        self.destinationRootAddressLabel.setText("")
        self.destinationRootAddressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationRootAddressLabel.setObjectName("destinationRootAddressLabel")
        self.GridLayout_2.addWidget(self.destinationRootAddressLabel, 2, 0, 1, 1)
        self.destinationRootAddressValueLabel = QtWidgets.QLabel(self.groupBox_2)
        self.destinationRootAddressValueLabel.setText("")
        self.destinationRootAddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.destinationRootAddressValueLabel.setObjectName("destinationRootAddressValueLabel")
        self.GridLayout_2.addWidget(self.destinationRootAddressValueLabel, 2, 1, 1, 1)
        self.destinationDerivationValueLabel = QtWidgets.QLabel(self.groupBox_2)
        self.destinationDerivationValueLabel.setText("")
        self.destinationDerivationValueLabel.setObjectName("destinationDerivationValueLabel")
        self.GridLayout_2.addWidget(self.destinationDerivationValueLabel, 5, 1, 1, 1)
        self.destinationRootNameOrAddressComboBox = QtWidgets.QComboBox(self.groupBox_2)
        self.destinationRootNameOrAddressComboBox.setObjectName("destinationRootNameOrAddressComboBox")
        self.GridLayout_2.addWidget(self.destinationRootNameOrAddressComboBox, 1, 1, 1, 1)
        self.destinationBalanceLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationBalanceLabel.setFont(font)
        self.destinationBalanceLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationBalanceLabel.setObjectName("destinationBalanceLabel")
        self.GridLayout_2.addWidget(self.destinationBalanceLabel, 7, 0, 1, 1)
        self.destinationBalanceValueLabel = QtWidgets.QLabel(self.groupBox_2)
        self.destinationBalanceValueLabel.setText("")
        self.destinationBalanceValueLabel.setObjectName("destinationBalanceValueLabel")
        self.GridLayout_2.addWidget(self.destinationBalanceValueLabel, 7, 1, 1, 1)
        self.destinationNameLineEdit = QtWidgets.QLineEdit(self.groupBox_2)
        self.destinationNameLineEdit.setObjectName("destinationNameLineEdit")
        self.GridLayout_2.addWidget(self.destinationNameLineEdit, 8, 1, 1, 1)
        self.destinationDerivationLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationDerivationLabel.setFont(font)
        self.destinationDerivationLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationDerivationLabel.setObjectName("destinationDerivationLabel")
        self.GridLayout_2.addWidget(self.destinationDerivationLabel, 5, 0, 1, 1)
        self.destinationNameLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationNameLabel.setFont(font)
        self.destinationNameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationNameLabel.setObjectName("destinationNameLabel")
        self.GridLayout_2.addWidget(self.destinationNameLabel, 8, 0, 1, 1)
        self.destinationAddressValueLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationAddressValueLabel.setFont(font)
        self.destinationAddressValueLabel.setText("")
        self.destinationAddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.destinationAddressValueLabel.setObjectName("destinationAddressValueLabel")
        self.GridLayout_2.addWidget(self.destinationAddressValueLabel, 6, 1, 1, 1)
        self.destinationMnemonicLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationMnemonicLabel.setFont(font)
        self.destinationMnemonicLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationMnemonicLabel.setObjectName("destinationMnemonicLabel")
        self.GridLayout_2.addWidget(self.destinationMnemonicLabel, 3, 0, 1, 1)
        self.destinationMnemonicLanguageLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationMnemonicLanguageLabel.setFont(font)
        self.destinationMnemonicLanguageLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationMnemonicLanguageLabel.setObjectName("destinationMnemonicLanguageLabel")
        self.GridLayout_2.addWidget(self.destinationMnemonicLanguageLabel, 4, 0, 1, 1)
        self.destinationMnemonicLanguageComboBox = QtWidgets.QComboBox(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.destinationMnemonicLanguageComboBox.sizePolicy().hasHeightForWidth())
        self.destinationMnemonicLanguageComboBox.setSizePolicy(sizePolicy)
        self.destinationMnemonicLanguageComboBox.setObjectName("destinationMnemonicLanguageComboBox")
        self.GridLayout_2.addWidget(self.destinationMnemonicLanguageComboBox, 4, 1, 1, 1)
        self.verticalLayout_3.addLayout(self.GridLayout_2)
        self.verticalLayout_4.addWidget(self.groupBox_2)
        self.destinationErrorLabel = QtWidgets.QLabel(self.wizardPage3)
        self.destinationErrorLabel.setText("")
        self.destinationErrorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.destinationErrorLabel.setObjectName("destinationErrorLabel")
        self.verticalLayout_4.addWidget(self.destinationErrorLabel)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.storedPasswordLabel = QtWidgets.QLabel(self.wizardPage3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.storedPasswordLabel.sizePolicy().hasHeightForWidth())
        self.storedPasswordLabel.setSizePolicy(sizePolicy)
        self.storedPasswordLabel.setSizeIncrement(QtCore.QSize(0, 0))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.storedPasswordLabel.setFont(font)
        self.storedPasswordLabel.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.storedPasswordLabel.setObjectName("storedPasswordLabel")
        self.horizontalLayout.addWidget(self.storedPasswordLabel)
        self.storedPasswordLineEdit = QtWidgets.QLineEdit(self.wizardPage3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.storedPasswordLineEdit.sizePolicy().hasHeightForWidth())
        self.storedPasswordLineEdit.setSizePolicy(sizePolicy)
        self.storedPasswordLineEdit.setMaxLength(20)
        self.storedPasswordLineEdit.setEchoMode(QtWidgets.QLineEdit.Password)
        self.storedPasswordLineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.storedPasswordLineEdit.setReadOnly(True)
        self.storedPasswordLineEdit.setObjectName("storedPasswordLineEdit")
        self.horizontalLayout.addWidget(self.storedPasswordLineEdit)
        self.storedPasswordHintLabel = QtWidgets.QLabel(self.wizardPage3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.storedPasswordHintLabel.setFont(font)
        self.storedPasswordHintLabel.setObjectName("storedPasswordHintLabel")
        self.horizontalLayout.addWidget(self.storedPasswordHintLabel)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        importAccountV1Wizard.addPage(self.wizardPage3)
        self.wizardPage4 = QtWidgets.QWizardPage()
        self.wizardPage4.setObjectName("wizardPage4")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.wizardPage4)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label = QtWidgets.QLabel(self.wizardPage4)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_5.addWidget(self.label)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.sourceV1AddressValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceV1AddressValueImportLabel.sizePolicy().hasHeightForWidth())
        self.sourceV1AddressValueImportLabel.setSizePolicy(sizePolicy)
        self.sourceV1AddressValueImportLabel.setText("")
        self.sourceV1AddressValueImportLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.sourceV1AddressValueImportLabel.setObjectName("sourceV1AddressValueImportLabel")
        self.gridLayout_2.addWidget(self.sourceV1AddressValueImportLabel, 1, 1, 1, 1)
        self.sourceV1AddressImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceV1AddressImportLabel.sizePolicy().hasHeightForWidth())
        self.sourceV1AddressImportLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceV1AddressImportLabel.setFont(font)
        self.sourceV1AddressImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceV1AddressImportLabel.setObjectName("sourceV1AddressImportLabel")
        self.gridLayout_2.addWidget(self.sourceV1AddressImportLabel, 1, 0, 1, 1)
        self.sourceBalanceValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceBalanceValueImportLabel.sizePolicy().hasHeightForWidth())
        self.sourceBalanceValueImportLabel.setSizePolicy(sizePolicy)
        self.sourceBalanceValueImportLabel.setText("")
        self.sourceBalanceValueImportLabel.setObjectName("sourceBalanceValueImportLabel")
        self.gridLayout_2.addWidget(self.sourceBalanceValueImportLabel, 3, 1, 1, 1)
        self.sourceBalanceImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceBalanceImportLabel.setFont(font)
        self.sourceBalanceImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceBalanceImportLabel.setObjectName("sourceBalanceImportLabel")
        self.gridLayout_2.addWidget(self.sourceBalanceImportLabel, 3, 0, 1, 1)
        self.sourceIdentityImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceIdentityImportLabel.setFont(font)
        self.sourceIdentityImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceIdentityImportLabel.setObjectName("sourceIdentityImportLabel")
        self.gridLayout_2.addWidget(self.sourceIdentityImportLabel, 4, 0, 1, 1)
        self.sourceIdentityValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceIdentityValueImportLabel.sizePolicy().hasHeightForWidth())
        self.sourceIdentityValueImportLabel.setSizePolicy(sizePolicy)
        self.sourceIdentityValueImportLabel.setText("")
        self.sourceIdentityValueImportLabel.setObjectName("sourceIdentityValueImportLabel")
        self.gridLayout_2.addWidget(self.sourceIdentityValueImportLabel, 4, 1, 1, 1)
        self.sourceAddressImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.sourceAddressImportLabel.setFont(font)
        self.sourceAddressImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sourceAddressImportLabel.setObjectName("sourceAddressImportLabel")
        self.gridLayout_2.addWidget(self.sourceAddressImportLabel, 2, 0, 1, 1)
        self.sourceAddressValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceAddressValueImportLabel.sizePolicy().hasHeightForWidth())
        self.sourceAddressValueImportLabel.setSizePolicy(sizePolicy)
        self.sourceAddressValueImportLabel.setText("")
        self.sourceAddressValueImportLabel.setObjectName("sourceAddressValueImportLabel")
        self.gridLayout_2.addWidget(self.sourceAddressValueImportLabel, 2, 1, 1, 1)
        self.verticalLayout_5.addLayout(self.gridLayout_2)
        self.importButton = QtWidgets.QPushButton(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.importButton.sizePolicy().hasHeightForWidth())
        self.importButton.setSizePolicy(sizePolicy)
        self.importButton.setObjectName("importButton")
        self.verticalLayout_5.addWidget(self.importButton, 0, QtCore.Qt.AlignHCenter)
        self.ImportImagelabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ImportImagelabel.sizePolicy().hasHeightForWidth())
        self.ImportImagelabel.setSizePolicy(sizePolicy)
        self.ImportImagelabel.setMaximumSize(QtCore.QSize(64, 64))
        self.ImportImagelabel.setText("")
        self.ImportImagelabel.setPixmap(QtGui.QPixmap(":/images/arrow_down"))
        self.ImportImagelabel.setScaledContents(True)
        self.ImportImagelabel.setAlignment(QtCore.Qt.AlignCenter)
        self.ImportImagelabel.setObjectName("ImportImagelabel")
        self.verticalLayout_5.addWidget(self.ImportImagelabel, 0, QtCore.Qt.AlignHCenter)
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.destinationAddressImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationAddressImportLabel.setFont(font)
        self.destinationAddressImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationAddressImportLabel.setObjectName("destinationAddressImportLabel")
        self.gridLayout_3.addWidget(self.destinationAddressImportLabel, 1, 0, 1, 1)
        self.destinationAddressValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.destinationAddressValueImportLabel.sizePolicy().hasHeightForWidth())
        self.destinationAddressValueImportLabel.setSizePolicy(sizePolicy)
        self.destinationAddressValueImportLabel.setText("")
        self.destinationAddressValueImportLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.destinationAddressValueImportLabel.setObjectName("destinationAddressValueImportLabel")
        self.gridLayout_3.addWidget(self.destinationAddressValueImportLabel, 1, 1, 1, 1)
        self.destinationBalanceValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        self.destinationBalanceValueImportLabel.setText("")
        self.destinationBalanceValueImportLabel.setObjectName("destinationBalanceValueImportLabel")
        self.gridLayout_3.addWidget(self.destinationBalanceValueImportLabel, 2, 1, 1, 1)
        self.destinationBalanceImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationBalanceImportLabel.setFont(font)
        self.destinationBalanceImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationBalanceImportLabel.setObjectName("destinationBalanceImportLabel")
        self.gridLayout_3.addWidget(self.destinationBalanceImportLabel, 2, 0, 1, 1)
        self.destinationIdentityImportLabel = QtWidgets.QLabel(self.wizardPage4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.destinationIdentityImportLabel.setFont(font)
        self.destinationIdentityImportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.destinationIdentityImportLabel.setObjectName("destinationIdentityImportLabel")
        self.gridLayout_3.addWidget(self.destinationIdentityImportLabel, 3, 0, 1, 1)
        self.destinationIdentityValueImportLabel = QtWidgets.QLabel(self.wizardPage4)
        self.destinationIdentityValueImportLabel.setText("")
        self.destinationIdentityValueImportLabel.setObjectName("destinationIdentityValueImportLabel")
        self.gridLayout_3.addWidget(self.destinationIdentityValueImportLabel, 3, 1, 1, 1)
        self.verticalLayout_5.addLayout(self.gridLayout_3)
        self.importErrorLabel = QtWidgets.QLabel(self.wizardPage4)
        self.importErrorLabel.setText("")
        self.importErrorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.importErrorLabel.setObjectName("importErrorLabel")
        self.verticalLayout_5.addWidget(self.importErrorLabel)
        importAccountV1Wizard.addPage(self.wizardPage4)

        self.retranslateUi(importAccountV1Wizard)
        QtCore.QMetaObject.connectSlotsByName(importAccountV1Wizard)
        importAccountV1Wizard.setTabOrder(self.sourceSecretIDLineEdit, self.sourcePasswordIDLineEdit)
        importAccountV1Wizard.setTabOrder(self.sourcePasswordIDLineEdit, self.sourceShowButton)
        importAccountV1Wizard.setTabOrder(self.sourceShowButton, self.destinationRootNameOrAddressComboBox)
        importAccountV1Wizard.setTabOrder(self.destinationRootNameOrAddressComboBox, self.destinationMnemonicLineEdit)
        importAccountV1Wizard.setTabOrder(self.destinationMnemonicLineEdit, self.destinationMnemonicLanguageComboBox)
        importAccountV1Wizard.setTabOrder(self.destinationMnemonicLanguageComboBox, self.destinationNameLineEdit)
        importAccountV1Wizard.setTabOrder(self.destinationNameLineEdit, self.destinationShowButton)
        importAccountV1Wizard.setTabOrder(self.destinationShowButton, self.storedPasswordLineEdit)
        importAccountV1Wizard.setTabOrder(self.storedPasswordLineEdit, self.importButton)

    def retranslateUi(self, importAccountV1Wizard):
        #  _translate = QtCore.QCoreApplication.translate
        importAccountV1Wizard.setWindowTitle(_("Wizard to import V1 account in V2 account"))
        self.page1_label.setText(_("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">This wizard will help you to import your V1 account into a V2 account.</p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">The V2 account to receive the identity and/or currency of the V1 account must exist in Tikka.</p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">If not, exit this wizard and create a V2 account by carefully and secretly writing down the restoration code (mnemonic).</p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; font-style:italic;\">V1 account: </span><span style=\" font-style:italic;\">A V1 account is an Ğ1 account generated by a secret ID and a password (with Cesium V1).</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-style:italic;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; font-style:italic;\">V2 Account:</span><span style=\" font-style:italic;\"> A V2 account is an account generated by a twelve-word restoration code (also called mnemonic).</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.page2_label.setText(_("Please enter the credentials of the V1 source account."))
        self.groupBox.setTitle(_("V1 Source Account"))
        self.sourceV1AddressLabel.setText(_("V1 Address"))
        self.sourceShowButton.setText(_("Show"))
        self.sourceAddressLabel.setText(_("Address"))
        self.sourcePasswordIDLabel.setText(_("Password"))
        self.sourceSecretIDLabel.setText(_("Secret ID"))
        self.sourceIdentityLabel.setText(_("Identity"))
        self.sourceBalanceLabel.setText(_("Balance"))
        self.page3_label.setText(_("Select a V2 destination account already created and confirm with the restoration mnemonic."))
        self.groupBox_2.setTitle(_("V2 Destination Account"))
        self.destinationShowButton.setText(_("Show"))
        self.destinationAddressLabel.setText(_("Address"))
        self.destinationRootNameOrAddressLabel.setText(_("Root account"))
        self.destinationBalanceLabel.setText(_("Balance"))
        self.destinationDerivationLabel.setText(_("Derivation"))
        self.destinationNameLabel.setText(_("Name"))
        self.destinationMnemonicLabel.setText(_("Mnemonic"))
        self.destinationMnemonicLanguageLabel.setText(_("Mnemonic language"))
        self.storedPasswordLabel.setText(_("Password"))
        self.storedPasswordHintLabel.setText(_("Wallet will be locked with the current password"))
        self.label.setText(_("Click on the import button and wait for the status message. The V1 source account will be deleted from the blockchain on success."))
        self.sourceV1AddressImportLabel.setText(_("V1 Address"))
        self.sourceBalanceImportLabel.setText(_("Balance"))
        self.sourceIdentityImportLabel.setText(_("Identity"))
        self.sourceAddressImportLabel.setText(_("Address"))
        self.importButton.setText(_("Import"))
        self.destinationAddressImportLabel.setText(_("Address"))
        self.destinationBalanceImportLabel.setText(_("Balance"))
        self.destinationIdentityImportLabel.setText(_("Identity"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5 import QtWidgets
from PyQt5.QtCore import QModelIndex, QMutex, QPoint, Qt
from PyQt5.QtWidgets import QApplication, QMainWindow, QStyleOptionViewItem, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import AccountEvent, CurrencyEvent, UnitEvent
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.slots.pyqt.entities.constants import (
    ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.models.account_table import AccountTableModel
from tikka.slots.pyqt.resources.gui.widgets.account_table_rc import (
    Ui_AccountTableWidget,
)
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu


class IconDelegate(QtWidgets.QStyledItemDelegate):
    """
    IconDelegate class to center icons in table view
    """

    def initStyleOption(self, option: QStyleOptionViewItem, index: QModelIndex):
        """
        Init Style

        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        super().initStyleOption(option, index)
        # resize width to actual width of icon
        option.decorationSize.setWidth(option.rect.width())


class AccountTableWidget(QWidget, Ui_AccountTableWidget):
    """
    AccountTableWidget class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountTableWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.mutex = mutex
        self.account_table_model = AccountTableModel(self.application, self.locale())
        self._ = self.application.translator.gettext

        # use icon delegate to center icons
        delegate = IconDelegate(self.tableView)
        self.tableView.setItemDelegate(delegate)

        # setup table view
        self.tableView.setModel(self.account_table_model)
        # set header sort column and sort order icon
        sort_column_preference = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if sort_column_preference is not None:
            sort_column_index = int(sort_column_preference)
        else:
            sort_column_index = -1
        repository_sort_order = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        self.tableView.horizontalHeader().setSortIndicator(
            sort_column_index,
            Qt.AscendingOrder
            if repository_sort_order == AccountsRepositoryInterface.SORT_ORDER_ASCENDING
            else Qt.DescendingOrder,
        )
        self.tableView.setSortingEnabled(True)
        self.init_category_filter()
        self.init_wallet_filter()
        self.tableView.resizeColumnsToContents()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_from_network_async_qworker = AsyncQWorker(
            self.fetch_from_network, self.mutex
        )
        self.fetch_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_from_network
        )

        # events
        self.refreshButton.clicked.connect(self.fetch_from_network_async_qworker.start)

        # table events
        self.tableView.customContextMenuRequested.connect(self.on_context_menu)
        self.tableView.model().dataChanged.connect(self.on_data_changed)
        self.categoryComboBox.currentIndexChanged.connect(
            self.on_category_filter_changed
        )
        self.walletComboBox.currentIndexChanged.connect(self.on_wallet_filter_changed)
        # application events
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, self.on_unit_event
        )
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self.on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_ADD, self.on_add_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_DELETE, self.on_delete_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, self.on_update_account_event
        )

    def init_category_filter(self):
        """
        Fill category filter combo box with all categories and a blank choice

        :return:
        """
        # init combo box
        self.categoryComboBox.addItem(self._("No filter"), userData=None)
        for category in self.application.categories.list_all():
            self.categoryComboBox.addItem(category.name, userData=category.id.hex)

        # set current category filter from preferences
        category_id = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY
        )
        self.categoryComboBox.setCurrentIndex(
            self.categoryComboBox.findData(category_id)
        )

    def on_category_filter_changed(self, _):
        """
        Update preferences with category filter for table view

        :return:
        """
        category_id = self.categoryComboBox.currentData()
        self.application.preferences_repository.set(
            ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY,
            None if category_id is None else str(category_id),
        )
        self._update_model()

    def init_wallet_filter(self):
        """
        Fill wallet filter combo box with None (with and without), True (with) or False (without)

        :return:
        """
        # init combo box
        self.walletComboBox.addItem(self._("No filter"), userData=None)
        self.walletComboBox.addItem(self._("Yes"), userData=True)
        self.walletComboBox.addItem(self._("No"), userData=False)

        # set current category filter from preferences
        wallet_filter_preference = self.application.preferences_repository.get(
            ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY
        )
        wallet_filter = None
        if wallet_filter_preference is not None:
            # preference store boolean as integer in a string column ("0" or "1")
            # convert it to boolean
            wallet_filter = int(wallet_filter_preference) == 1
        self.walletComboBox.setCurrentIndex(self.walletComboBox.findData(wallet_filter))

    def on_wallet_filter_changed(self, _):
        """
        Update preferences with wallet filter for table view

        :return:
        """
        wallet_filter = self.walletComboBox.currentData()
        self.application.preferences_repository.set(
            ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY, wallet_filter
        )
        self._update_model()

    def on_data_changed(self, _):
        """
        Triggered when data changed in table model

        :return:
        """
        self._update_model()

    def _update_model(self):
        """
        Update all data in model

        :return:
        """
        self.tableView.model().init_data()
        self.tableView.resizeColumnsToContents()

    def on_unit_event(self, _):
        """
        When a unit event is triggered

        :return:
        """
        self._update_model()

    def on_currency_event(self, _):
        """
        When a currency event is triggered

        :return:
        """
        self._update_model()

    def on_add_account_event(self, _):
        """
        Add account row when account is created

        :return:
        """
        self._update_model()

    def on_delete_account_event(self, _):
        """
        Remove account row when account is deleted

        :return:
        """
        self._update_model()

    def on_update_account_event(self, _):
        """
        Update account row when account is updated

        :return:
        """
        self._update_model()

    def on_context_menu(self, position: QPoint):
        """
        When right button on table widget

        :param position: QPoint instance
        :return:
        """
        # get selected account
        table_view_row = self.tableView.currentIndex().internalPointer()
        account = self.application.accounts.get_by_address(table_view_row.address)
        if account is not None:
            # display popup menu at click position
            AccountPopupMenu(self.application, account, self.mutex, self).exec_(
                self.tableView.mapToGlobal(position)
            )

    def fetch_from_network(self):
        """
        Fetch table model accounts data from the network

        :return:
        """
        self.refreshButton.setEnabled(False)
        self.errorLabel.setText("")

        try:
            self.application.accounts.network_update_balances(
                self.application.accounts.get_list()
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

        try:
            identity_indice = [
                identity_index
                for identity_index in self.application.identities.network_get_identity_indice(
                    [
                        account.address
                        for account in self.application.accounts.get_list()
                    ]
                )
                if identity_index is not None
            ]
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
        else:
            try:
                self.application.identities.network_get_identities(identity_indice)
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
            else:
                try:
                    self.application.smiths.network_get_smiths(identity_indice)
                except Exception as exception:
                    self.errorLabel.setText(self._(str(exception)))

    def _on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.refreshButton.setEnabled(True)
        self._update_model()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(
        AccountTableWidget(application_, QMutex(), main_window)
    )

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import shutil
from pathlib import Path

from tikka.domains.entities.constants import CONFIG_FILENAME
from tikka.interfaces.adapters.repository.config import ConfigRepositoryInterface

ASSETS_PATH = Path(__file__).parent.joinpath("assets")
DEFAULT_CONFIG_PATH = ASSETS_PATH.joinpath(CONFIG_FILENAME)


class FileConfigRepository(ConfigRepositoryInterface):
    """
    FileConfigRepository class
    """

    filepath = None

    def __init__(self, path: Path):
        """
        Create config file in user config path

        :param path: Path instance of config file path
        :return:
        """
        self.filepath = Path().joinpath(path, CONFIG_FILENAME).expanduser()

        if not path.expanduser().exists():
            os.makedirs(path.expanduser())

        if not self.filepath.exists():
            # copy default config in user config path
            shutil.copyfile(DEFAULT_CONFIG_PATH.expanduser(), self.filepath)

    def load(self):
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConfigRepositoryInterface.load.__doc__
        )
        with DEFAULT_CONFIG_PATH.expanduser().open(
            "r", encoding="utf-8"
        ) as file_handler:
            default_data = json.load(file_handler)

        with self.filepath.open("r", encoding="utf-8") as file_handler:
            data = json.load(file_handler)

        data_keys_updated = False
        # add new config key if default config file updated
        for key in default_data:
            if key not in data:
                data[key] = default_data[key]
                data_keys_updated = True
        data_keys_to_delete = []
        # purge obsolete keys if default config file updated
        for key in data:
            if key not in default_data:
                data_keys_to_delete.append(key)
                data_keys_updated = True
        for key in data_keys_to_delete:
            del data[key]

        if data_keys_updated is True:
            self.save(data)

        return data

    def save(self, data: dict):
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConfigRepositoryInterface.save.__doc__
        )
        if self.filepath is not None:
            with self.filepath.open("w", encoding="utf-8") as file_handler:
                json.dump(data, file_handler)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List, Optional

from tikka.domains.entities.smith import Smith


class SmithsRepositoryInterface(abc.ABC):
    """
    SmithsRepositoryInterface class
    """

    @abc.abstractmethod
    def add(self, smith: Smith) -> None:
        """
        Add a new smith in repository

        :param smith: Smith instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, index: int) -> Optional[Smith]:
        """
        Return Smith instance from repository

        :param index: Identity index
        :return:
        """
        raise NotImplementedError

    def update(self, smith: Smith) -> None:
        """
        Update smith in repository

        :param smith: Smith instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, index: int) -> None:
        """
        Delete smith in repository

        :param index: Identity index of smith to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, index: int) -> bool:
        """
        Return True if smith with identity index is in repository, else False

        :param index: Identity index of smith to check
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(self) -> List[Smith]:
        """
        Return list of all smiths in repository

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from tikka.adapters.network.rpc.connection import RPCConnection
from tikka.domains.entities.events import ConnectionsEvent
from tikka.domains.entities.node import Node
from tikka.domains.events import EventDispatcher
from tikka.interfaces.domains.connections import ConnectionsInterface


class Connections(ConnectionsInterface):
    """
    Connections class
    """

    def __init__(self, rpc: RPCConnection, event_dispatcher: EventDispatcher) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionsInterface.__init__.__doc__
        )
        super().__init__(rpc, event_dispatcher)

    def connect(self, node: Node):
        """
        Connect all connections

        :return:
        """
        self.rpc.connect(node)
        if self.rpc.is_connected():
            self.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_CONNECTED)
            )
        else:
            self.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_DISCONNECTED)
            )

    def disconnect(self):
        """
        Disconnect all connections

        :return:
        """
        if self.rpc.is_connected():
            self.rpc.disconnect()
        self.event_dispatcher.dispatch_event(
            ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_DISCONNECTED)
        )

    def is_connected(self) -> bool:
        """
        Return True if connection is active, False otherwise

        :return:
        """
        return self.rpc.is_connected()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Dict, List

from substrateinterface import Keypair

from tikka.domains.entities.authorities import AuthorityStatus
from tikka.interfaces.domains.connections import ConnectionsInterface


class NetworkAuthoritiesInterface(abc.ABC):
    """
    NetworkAuthoritiesInterface class
    """

    def __init__(self, connections: ConnectionsInterface) -> None:
        """
        Use connections to request/send online authorities information

        :param connections: ConnectionsInterface instance
        :return:
        """
        self.connections = connections

    @abc.abstractmethod
    def rotate_keys(self) -> str:
        """
        Rotate Session keys

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def has_session_keys(self, session_keys: str) -> bool:
        """
        Return True if the current node keystore store private session keys corresponding to public session_keys

        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def publish_session_keys(self, keypair: Keypair, session_keys: str) -> None:
        """
        Set/Change in blockchain the session public keys for the Keypair account

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def go_online(self, keypair: Keypair) -> None:
        """
        Start writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def go_offline(self, keypair: Keypair) -> None:
        """
        Stop writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_status(self, identity_index: int) -> AuthorityStatus:
        """
        Return AuthorityStatus of identity_index

        :param identity_index: Identity index of Smith to get status of
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_all(self) -> Dict[int, List[int]]:
        """
        Return a dict of list of all identity index in incoming authorities

        {
            AuthorityStatus.INCOMING.value: [identity_indice,...],
            AuthorityStatus.ONLINE.value: [identity_indice,...],
            AuthorityStatus.OUTGOING.value": [identity_indice,...],
        }

        :return:
        """
        raise NotImplementedError


class NetworkAuthoritiesException(Exception):
    """
    NetworkAuthoritiesException
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QApplication, QMenu, QMessageBox, QWidget
from substrateinterface import KeypairType

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.windows.account_derivation_create import (
    AccountDerivationCreateWindow,
)
from tikka.slots.pyqt.windows.account_rename import AccountRenameWindow
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow
from tikka.slots.pyqt.windows.transfer import TransferWindow
from tikka.slots.pyqt.windows.v1_wallet_restore import V1WalletRestoreWindow
from tikka.slots.pyqt.windows.v1_wallet_restore_from_file import (
    V1WalletRestoreFromFileWindow,
)
from tikka.slots.pyqt.windows.wallet_restore import WalletRestoreWindow


class AccountPopupMenu(QMenu):
    """
    AccountPopupMenu class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ):
        """
        Init AccountPopupMenu instance

        :param application: Application instance
        :param account: Account instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)

        self.application = application
        self.account = account
        self.mutex = mutex
        self._ = self.application.translator.gettext

        transfer_to_action = self.addAction(self._("Transfer to"))
        transfer_to_action.triggered.connect(self.transfer_to)
        if self.application.wallets.exists(account.address):
            transfer_from_action = self.addAction(self._("Transfer from"))
            transfer_from_action.triggered.connect(self.transfer_from)

        # menu actions
        copy_address_to_clipboard_action = self.addAction(
            self._("Copy address to clipboard")
        )
        copy_address_to_clipboard_action.triggered.connect(
            self.copy_address_to_clipboard
        )
        if self.application.wallets.exists(account.address):
            if self.application.wallets.is_unlocked(account.address) is False:
                unlock_account_action = self.addAction(self._("Unlock account access"))
                unlock_account_action.triggered.connect(self.unlock_account)
            else:
                lock_account_action = self.addAction(self._("Lock account access"))
                lock_account_action.triggered.connect(self.lock_account)

            wallet_forget_action = self.addAction(self._("Forget the wallet"))
            wallet_forget_action.triggered.connect(self.wallet_forget)
        else:
            wallet_restore_action = self.addAction(self._("Store the wallet"))
            wallet_restore_action.triggered.connect(self.wallet_restore)

        # if account is a root account...
        if self.account.root is None:
            wallet_password_forgotten_action = self.addAction(
                self._("Password forgotten / change")
            )
            wallet_password_forgotten_action.triggered.connect(
                self.wallet_password_forgotten
            )
            # derivation only for SR25519 accounts...
            if self.account.crypto_type == KeypairType.SR25519:
                create_derived_account_action = self.addAction(
                    self._("Create derived account")
                )
                create_derived_account_action.triggered.connect(
                    self.create_derived_account
                )

        rename_account_action = self.addAction(self._("Rename account"))
        rename_account_action.triggered.connect(self.rename_account)

        forget_account_action = self.addAction(self._("Forget account"))
        forget_account_action.triggered.connect(self.confirm_forget_account)

    def copy_address_to_clipboard(self):
        """
        Copy address of selected row to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.account.address)

    def unlock_account(self):
        """
        Open account unlock window

        :return:
        """
        AccountUnlockWindow(self.application, self.account, self).exec_()

    def transfer_to(self):
        """
        Open transfer window with account as recipient

        :return:
        """
        TransferWindow(
            self.application, self.mutex, None, self.account, parent=self
        ).exec_()

    def transfer_from(self):
        """
        Open transfer window with account as sender

        :return:
        """
        TransferWindow(
            self.application, self.mutex, self.account, None, parent=self
        ).exec_()

    def lock_account(self):
        """
        Lock account

        :return:
        """
        self.application.accounts.lock(self.account)

    def wallet_password_forgotten(self):
        """
        Open wallet restore window to reset wallet password

        :return:
        """
        # if wallet type is V2...
        if self.account.crypto_type == KeypairType.SR25519:
            WalletRestoreWindow(self.application, self.account, True).exec_()
        # wallet V1
        else:
            if self.account.file_import is True:
                V1WalletRestoreFromFileWindow(
                    self.application, self.account, True
                ).exec_()
            else:
                V1WalletRestoreWindow(self.application, self.account, True).exec_()

    def wallet_restore(self):
        """
        Open wallet restore window

        :return:
        """
        # if wallet type is V2...
        if self.account.crypto_type == KeypairType.SR25519:
            WalletRestoreWindow(self.application, self.account).exec_()
        # wallet V1
        else:
            if self.account.file_import is True:
                V1WalletRestoreFromFileWindow(self.application, self.account).exec_()
            else:
                V1WalletRestoreWindow(self.application, self.account).exec_()

    def wallet_forget(self):
        """
        Forget wallet for this account

        :return:
        """
        # display confirm dialog and get response
        button = QMessageBox.question(
            self,
            self._("Forget wallet"),
            self._("Forget wallet for account {address}?").format(
                address=self.account.address
            ),
        )
        if button == QMessageBox.Yes:
            self.application.accounts.forget_wallet(self.account)

    def create_derived_account(self):
        """
        Open create account derivation window

        :return:
        """
        AccountDerivationCreateWindow(self.application, self.account).exec_()

    def rename_account(self):
        """
        Open rename account window

        :return:
        """
        AccountRenameWindow(self.application, self.account).exec_()

    def confirm_forget_account(self):
        """
        Display confirm dialog then forget account if confirmed

        :return:
        """
        # display confirm dialog and get response
        button = QMessageBox.question(
            self,
            self._("Forget account"),
            self._("Forget account {address} (and derived accounts if any)?").format(
                address=self.account.address
            ),
        )
        if button == QMessageBox.Yes:
            self.application.accounts.delete(self.account)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = Account("732SSfuwjB7jkt9th1zerGhphs6nknaCBCTozxUcPWPU")

    menu = AccountPopupMenu(application_, account_, QMutex())
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable

from tikka.interfaces.entities.events import EventInterface


class EventDispatcher:
    """
    Generic event dispatcher which listen and dispatch events
    """

    def __init__(self) -> None:
        """
        Init EventDispatcher instance

        :return:
        """
        self._events: dict = {}

    def __del__(self):
        """
        Remove all listener references at destruction time

        :return:
        """
        self._events = None

    def has_listener(self, event_type: str, listener: Callable) -> bool:
        """
        Return true if listener is registered to event_type

        :param event_type: Type of event
        :param listener: Callable method or function to call when event occurs
        :return:
        """
        # Check for event type and for the listener
        if event_type in self._events:
            return listener in self._events[event_type]

        return False

    def dispatch_event(self, event: EventInterface) -> None:
        """
        Dispatch an instance of Event class

        :param event: Event instance
        :return:
        """
        # Dispatch the event to all the associated listeners
        if event.type in self._events:
            listeners = self._events[event.type]

            for listener in listeners:
                listener(event)

    def add_event_listener(self, event_type: str, listener: Callable) -> None:
        """
        Add an event listener for an event type

        :param event_type: Type of event
        :param listener: Callable method or function to call when event occurs
        :return:
        """
        # Add listener to the event type
        if not self.has_listener(event_type, listener):
            listeners = self._events.get(event_type, [])

            listeners.append(listener)

            self._events[event_type] = listeners

    def remove_event_listener(self, event_type: str, listener: Callable) -> None:
        """
        Remove event listener.

        :param event_type: Type of event
        :param listener: Callable method or function to call when event occurs
        :return:
        """
        # Remove the listener from the event type
        if self.has_listener(event_type, listener):
            listeners = self._events[event_type]

            if len(listeners) == 1:
                # Only this listener remains so remove the key
                del self._events[event_type]

            else:
                # Update listeners chain
                listeners.remove(listener)

                self._events[event_type] = listeners



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from pathlib import Path
from typing import Optional, Union

from tikka.domains.entities.wallet import V1FileWallet


class V1FileWalletsRepositoryInterface(abc.ABC):
    """
    V1FileWalletsRepositoryInterface class
    """

    @abc.abstractmethod
    def load(
        self, path: Union[str, Path], password: Optional[str] = None
    ) -> V1FileWallet:
        """
        Return data Wallet instance

        :param path: Path instance or string of the file
        :param password: Password for encrypted file
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import sys
from typing import Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import UnitEvent
from tikka.slots.pyqt.entities.constants import SELECTED_UNIT_PREFERENCES_KEY
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.widgets.currency_rc import Ui_currencyWidget


class CurrencyWidget(QWidget, Ui_currencyWidget):
    """
    CurrencyWidget class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ):
        """
        Init currency widget

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex
        self._update_ui()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_from_network_async_qworker = AsyncQWorker(
            self.fetch_from_network, self.mutex
        )
        self.fetch_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_from_network
        )

        # events
        self.refreshButton.clicked.connect(self.fetch_from_network_async_qworker.start)

        # application events
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, lambda e: self._update_ui()
        )

        if self.application.connections.is_connected():
            self.fetch_from_network_async_qworker.start()

    def fetch_from_network(self):
        """
        Fetch last currency data from the network

        :return:
        """
        self.errorLabel.setText("")
        self.refreshButton.setEnabled(False)
        try:
            self.application.currencies.network_get_properties()
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

    def _on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.refreshButton.setEnabled(True)
        self._update_ui()

    def _update_ui(self):
        """
        Update UI values

        :return:
        """
        unit_amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)
        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            selected_amount = self.application.amounts.get_amount(unit_preference)
        else:
            selected_amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # Currency parameters
        currency = self.application.currencies.get_current()
        self.nameValueLabel.setText(currency.name)
        self.symbolValueLabel.setText(currency.token_symbol)

        self.universalDividendValueLabel.setText(
            ""
            if currency.universal_dividend is None
            else self.locale().toCurrencyString(
                unit_amount.value(currency.universal_dividend), unit_amount.symbol()
            )
        )
        self.monetaryMassValueLabel.setText(
            ""
            if currency.monetary_mass is None
            else self.locale().toCurrencyString(
                selected_amount.value(currency.monetary_mass), selected_amount.symbol()
            )
        )
        self.membersValueLabel.setText(
            ""
            if currency.members_count is None
            else self.locale().toString(currency.members_count)
        )

        # Blockchain parameters
        self.expectedBlockDurationValueLabel.setText(
            ""
            if currency.block_duration is None
            else str(datetime.timedelta(milliseconds=currency.block_duration))
        )

        self.expectedEpochDurationValueLabel.setText(
            ""
            if currency.epoch_duration is None
            else str(datetime.timedelta(milliseconds=currency.epoch_duration))
        )

    def _on_refresh_button_clicked_event(self):
        """
        Triggered when user click on refresh button

        :return:
        """
        # Disable button
        self.refreshButton.setEnabled(False)
        # Start the thread
        self.fetch_from_network_async_qworker.start()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(CurrencyWidget(application_, QMutex(), main_window))

    sys.exit(qapp.exec_())
